from g2p_fa.model import (Encoder, Decoder, Seq2Seq)
from g2p_fa.hparams import hparams as hp
from g2p_fa.utils import (IPA2tensor, tensor2IPA, word2tensor, WordDataset, MODEL_PATH)
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader

class G2P_Fa:
    def __init__(self, checkpoint=MODEL_PATH, hparams={}):
        for key, value in hparams.items():
            hp[key] = value

        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        
        enc = Encoder(hp['INPUT_DIM'], hp['ENC_EMB_DIM'], hp['HID_DIM'], hp['N_LAYERS'], hp['ENC_DROPOUT'])
        dec = Decoder(hp['OUTPUT_DIM'], hp['DEC_EMB_DIM'], hp['HID_DIM'], hp['N_LAYERS'], hp['DEC_DROPOUT'])

        self.model = Seq2Seq(enc, dec, self.device).to(self.device)
        if checkpoint:
            self.model.load_state_dict(torch.load(checkpoint, map_location=self.device))

    def train(self, data, epoch = 20, CLIP = 1):
        ds = WordDataset(data)
        train_len = int(len(ds) * 0.8)
        valid_len = len(ds) - train_len
        train_ds, valid_ds = torch.utils.data.random_split(ds, [train_len, valid_len])
        train_dataloader = DataLoader(train_ds, batch_size=1024, shuffle=True)
        valid_dataloader = DataLoader(valid_ds, batch_size=1024, shuffle=True)
        print(f'train samples: {len(train_ds)}, validation samples: {len(valid_ds)}')
        optimizer = optim.Adam(self.model.parameters())
        criterion = nn.CrossEntropyLoss(ignore_index=0)

        print(f'initial loss: {self.evaluate(valid_dataloader, criterion)}')
        for i in range(epoch):
            self.model.train()
            epoch_loss = 0
            for batch in train_dataloader:
                optimizer.zero_grad()

                src = batch[0].to(self.device)
                src = torch.transpose(src, 1, 0)
                trg = batch[1].to(self.device)
                trg = torch.transpose(trg, 1, 0)

                output = self.model(src, trg)
                output = output[1:].view(-1, hp['OUTPUT_DIM'])
                trg = trg[1:].reshape(-1)

                loss = criterion(output, trg)
        
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), CLIP)

                optimizer.step()

                epoch_loss += loss.item()

            eval_loss = self.evaluate(valid_dataloader, criterion)
            print(f'Epoch {i+1} / {epoch}\tTrain Loss: {epoch_loss/len(train_dataloader):.3f}, Valid loss: {eval_loss:.3f}')
        self.model.eval()

    def evaluate(self, iterator, criterion):
    
        self.model.eval()
        epoch_loss = 0
        with torch.no_grad():
            for i, batch in enumerate(iterator):

                src = batch[0].to(self.device)
                src = torch.transpose(src, 1, 0)
                trg = batch[1].to(self.device)
                trg = torch.transpose(trg, 1, 0)

                output = self.model(src, trg, 0) #turn off teacher forcing

                output_dim = output.shape[-1]
            
                output = output[1:].view(-1, output_dim)
                trg = trg[1:].reshape(-1)

                loss = criterion(output, trg)
            
                epoch_loss += loss.item()
        return epoch_loss / len(iterator)

    def __call__(self, word):
        word_vector = word2tensor(word).to(self.device)
        trg = IPA2tensor('b').to(self.device)
        output_vector = self.model(word_vector.unsqueeze(1), trg.unsqueeze(1), 0)
        output_word = tensor2IPA(output_vector.argmax(2))

        return output_word

    def save(self, PATH):
        torch.save(self.model.state_dict(), PATH)





import os
import torch
from torch.utils.data import Dataset

MAX_LEN = 30
THIS_DIR, _ = os.path.split(__file__)
MODEL_PATH = os.path.join(THIS_DIR, "data/weights")

class WordDataset(Dataset):
    def __init__(self, data_file):
        self.data = []
        with open(data_file, 'r', encoding='utf-8') as f:
            for line in f:
                word, IPA = line.strip().split(',')
                self.data.append((word2tensor(word, MAX_LEN),IPA2tensor(IPA, MAX_LEN)))

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]

FA_LETTERS = 'ءآئابتثجحخدذرزسشصضطظعغفقلمنهوُِپچژکگی'
IPA_LETTERS = 'bdefhijklmnopqstuvxzæɒɡɾʃʒʔʰː'
PAD = 0
START = 1
END = 2

def fa_letter2tensor(letter):
    return FA_LETTERS.find(letter) + 3

def IPA_letter2tensor(letter):
    return IPA_LETTERS.find(letter) + 3

def word2tensor(word, max_len=MAX_LEN):
    word_tensor = torch.zeros(max_len, dtype=torch.long)
    word_tensor[0] = START
    for i, letter in enumerate(word):
        word_tensor[i+1] = fa_letter2tensor(letter)
    word_tensor[len(word)+1] = END
    return word_tensor

def IPA2tensor(IPA, max_len=MAX_LEN):
    IPA_tensor = torch.zeros(max_len, dtype=torch.long)
    IPA_tensor[0] = START
    for i, letter in enumerate(IPA):
        IPA_tensor[i+1] = IPA_letter2tensor(letter)
    IPA_tensor[len(IPA)+1] = END
    return IPA_tensor

def tensor2IPA(tensor):
    idx = 1
    word = ''
    while(tensor[idx].item() != END):
        word += IPA_LETTERS[tensor[idx].item()-3]
        idx += 1
    return word
    




import torch
import torch.nn as nn

import random

class Encoder(nn.Module):
    def __init__(self, input_dim, emb_dim, hid_dim, n_layers, dropout):
        super().__init__()
        
        self.hid_dim = hid_dim
        self.n_layers = n_layers
        self.embedding = nn.Embedding(input_dim, emb_dim)
        self.rnn = nn.LSTM(emb_dim, hid_dim, n_layers, dropout = dropout)
        self.dropout = nn.Dropout(dropout)
        
    def forward(self, src):
        
        embedded = self.dropout(self.embedding(src))
        outputs, (hidden, cell) = self.rnn(embedded)
        return hidden, cell

class Decoder(nn.Module):
    def __init__(self, output_dim, emb_dim, hid_dim, n_layers, dropout):
        super().__init__()
        
        self.output_dim = output_dim
        self.hid_dim = hid_dim
        self.n_layers = n_layers
        
        self.embedding = nn.Embedding(output_dim, emb_dim)
        
        self.rnn = nn.LSTM(emb_dim, hid_dim, n_layers, dropout = dropout)
        
        self.fc_out = nn.Linear(hid_dim, output_dim)
        
        self.dropout = nn.Dropout(dropout)
        
    def forward(self, input, hidden, cell):
        
        input = input.unsqueeze(0)
        embedded = self.dropout(self.embedding(input))
        output, (hidden, cell) = self.rnn(embedded, (hidden, cell))
        prediction = self.fc_out(output.squeeze(0))
        return prediction, hidden, cell

class Seq2Seq(nn.Module):
    def __init__(self, encoder, decoder, device):
        super().__init__()
        
        self.encoder = encoder
        self.decoder = decoder
        self.device = device
        
        assert encoder.hid_dim == decoder.hid_dim, \
            "Hidden dimensions of encoder and decoder must be equal!"
        assert encoder.n_layers == decoder.n_layers, \
            "Encoder and decoder must have equal number of layers!"
        
    def forward(self, src, trg, teacher_forcing_ratio = 0.5):
        
        batch_size = trg.shape[1]
        trg_len = trg.shape[0]
        trg_vocab_size = self.decoder.output_dim
        
        outputs = torch.zeros(trg_len, batch_size, trg_vocab_size).to(self.device)
        
        hidden, cell = self.encoder(src)
        
        input = trg[0,:]
        
        for t in range(1, trg_len):
            
            output, hidden, cell = self.decoder(input, hidden, cell)
            
            outputs[t] = output
            
            teacher_force = random.random() < teacher_forcing_ratio
            
            top1 = output.argmax(1) 
            input = trg[t] if teacher_force else top1
        
        return outputs



import setuptools
from pathlib import Path

this_dir = Path(__file__).parent

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="g2p_fa",
    version="1.0.0",
    author="Mohamadhosein Dehghani",
    author_email="demh1377@gmail.com",
    description="A Persian Grapheme to Phoneme model using LSTM implemented in pytorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/de-mh/g2p_fa",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["g2p_fa"],
    install_requires=['torch'],
    package_data={'g2p_fa': ['data/*']},
    python_requires=">=3.6",
)